import java.util.*;
import java.io.*;

class BinSearchTree<T extends Comparable<? super T>> {
	T data;
	BinSearchTree<T> left,right,parent;
	int size=1;
	public BinSearchTree(T data){ this.data = data; this.parent = this;}
	public void add(T t){
		if(t==null) return;
		BinSearchTree<T> cur = this;
		while(true){
			cur.size++;
			if(t.compareTo(cur.data)<0){
				if(cur.left==null){
					cur.left = new BinSearchTree<T>(t); cur.left.parent = cur; return;
				}else{
					cur = cur.left;
				}
			}else{
				if(cur.right==null){
					cur.right = new BinSearchTree<T>(t);  cur.right.parent = cur; return;
				}else{
					cur = cur.right;
				}
			}
		}
	}
	public BinSearchTree<T> findMax(){
		BinSearchTree<T> cur = this;
		while(true){
			if(cur.right == null) return cur;
			cur = cur.right;
		}
	}
	public String toString(){
		return mytoString(0);
		//return (left!=null? "["+left.toString()+"] " :"") + data.toString() + "(" + size +")" + (right!=null? "["+right.toString()+"]" :"");
	}
	public String mytoString(int p){
		String prefix = "";
		for(int i=0; i<p; i++) prefix +="|-";
		String res = prefix+data.toString() + " ("+ size +")\n";
		if(left!=null) res += left.mytoString(p+1);
		if(right!=null) res += right.mytoString(p+1);
		return res;
	}
	public void printInOrder(){
		if(left != null){ System.out.print("( "); left.printInOrder(); System.out.print(" )"); }
		System.out.print(data);
		if(right != null){ System.out.print("( "); right.printInOrder(); System.out.print(" )"); }
	}
	public void printInOrderStack(){
		Deque<BinSearchTree<T>> myq = new ArrayDeque<>();
		BinSearchTree<T> tmp;
		myq.addLast(this);
		while(!myq.isEmpty()){
			tmp = myq.removeLast();
			if(tmp.left == null){
				System.out.print(" "+tmp.data);
				if(tmp.right!=null) myq.addLast(tmp.right);
			}else{
				if(tmp.right!=null) myq.addLast(tmp.right);
				myq.addLast(new BinSearchTree<T>(tmp.data));
				myq.addLast(tmp.left);
			}
		}
	}
	public void printInOrderLoop(){
		BinSearchTree<T> cur = this; BinSearchTree<T> prev = this;
		while(true){
			if(cur.left!=null && prev==cur.parent){ prev=cur; cur=cur.left; continue; }
			if(cur.left!=null && prev==cur.left) { System.out.print(" "+cur.data); }
			if(cur.left==null && prev==cur.parent) { System.out.print(" "+cur.data); }
			if(cur.right!=null && prev!=cur.right) { prev = cur; cur = cur.right; continue; }
			prev = cur; cur = cur.parent; 
			if(cur == prev) break;
		}
	}
	boolean contains(T t){
		if(data.equals(t)) return true;
		if(data.compareTo(t)<0 && right !=null) return right.contains(t);
		if(data.compareTo(t)>0 && left !=null) return left.contains(t);
		return false;
	}
	boolean containsLoop(T t){
		BinSearchTree<T> tmp = this;
		while(tmp!=null){
			if(tmp.data.equals(t)) return true;
			if(tmp.data.compareTo(t)<0) tmp = tmp.right;
			else if(tmp.data.compareTo(t)>0) tmp = tmp.left;
		}
		return false;
	}
}


class TestTree {
	public static void main(String [] args){
		BinSearchTree<Integer> t = new BinSearchTree<>(3);
		t.add(5);
		t.add(2);
		t.add(7);
		t.add(4);
		t.add(10); t.add(1); t.add(12);
		System.out.println(t);
		t.printInOrderStack(); System.out.println();
		for(int i=1; i<15; i++){
			System.out.println("t contains "+i+": "+t.contains(i));
		}
		for(int i=1; i<15; i++){
			System.out.println("t contains "+i+": "+t.containsLoop(i));
		}
	}
}
